import re
from typing import Collection, Optional

from fractal_specifications.generic.collections import AndSpecification, OrSpecification
from fractal_specifications.generic.operators import (
    EqualsSpecification,
    GreaterThanEqualSpecification,
    GreaterThanSpecification,
    InSpecification,
    LessThanEqualSpecification,
    LessThanSpecification,
    RegexStringMatchSpecification,
)
from fractal_specifications.generic.specification import (
    EmptySpecification,
    Specification,
)


class SpecificationNotMappedToMongo(Exception):
    pass


class MongoSpecificationBuilder:
    @staticmethod
    def build(specification: Optional[Specification] = None) -> Optional[Collection]:
        if specification is None:
            return None
        elif isinstance(specification, EmptySpecification):
            return None
        elif isinstance(specification, AndSpecification):
            specs = [
                s
                for spec in specification.to_collection()
                if (s := MongoSpecificationBuilder.build(spec))
            ]
            return {"$and": specs} if specs else None
        elif isinstance(specification, OrSpecification):
            specs = [
                s
                for spec in specification.to_collection()
                if (s := MongoSpecificationBuilder.build(spec))
            ]
            return {"$or": specs} if specs else None
        elif isinstance(specification, InSpecification):
            return {specification.field: {"$in": specification.value}}
        elif isinstance(specification, EqualsSpecification):
            return {specification.field: {"$eq": specification.value}}
        elif isinstance(specification, LessThanSpecification):
            return {specification.field: {"$lt": specification.value}}
        elif isinstance(specification, LessThanEqualSpecification):
            return {specification.field: {"$lte": specification.value}}
        elif isinstance(specification, GreaterThanSpecification):
            return {specification.field: {"$gt": specification.value}}
        elif isinstance(specification, GreaterThanEqualSpecification):
            return {specification.field: {"$gte": specification.value}}
        elif isinstance(specification, RegexStringMatchSpecification):
            return {
                specification.field: {"$regex": f".*{re.escape(specification.value)}.*"}
            }
        elif isinstance(specification.to_collection(), dict):
            specs = [
                {key: {"$eq": value}}
                for key, value in dict(specification.to_collection()).items()
            ]
            return {"$and": specs} if specs else None
        raise SpecificationNotMappedToMongo(
            f"Specification '{specification}' not mapped to Mongo query."
        )



from typing import Optional

from fractal_specifications.generic.collections import AndSpecification, OrSpecification
from fractal_specifications.generic.operators import (
    EqualsSpecification,
    GreaterThanEqualSpecification,
    GreaterThanSpecification,
    InSpecification,
    LessThanEqualSpecification,
    LessThanSpecification,
)
from fractal_specifications.generic.specification import (
    EmptySpecification,
    Specification,
)


class SpecificationNotMappedToElastic(Exception):
    pass


class ElasticSpecificationBuilder:
    @staticmethod
    def build(specification: Optional[Specification] = None) -> Optional[dict]:
        if specification is None:
            return None
        elif isinstance(specification, EmptySpecification):
            return None
        elif isinstance(specification, AndSpecification):
            return {
                "bool": {
                    "must": [
                        s
                        for spec in specification.to_collection()
                        if (s := ElasticSpecificationBuilder.build(spec))
                    ]
                }
            }
        elif isinstance(specification, OrSpecification):
            return {
                "bool": {
                    "should": [
                        s
                        for spec in specification.to_collection()
                        if (s := ElasticSpecificationBuilder.build(spec))
                    ]
                }
            }
        elif isinstance(specification, InSpecification):
            return {
                "query_string": {
                    "default_field": specification.field,
                    "query": specification.value,
                }
            }
        elif isinstance(specification, EqualsSpecification):
            return {"match": {"%s.keyword" % specification.field: specification.value}}
        elif isinstance(specification, LessThanSpecification):
            return {
                "bool": {
                    "filter": [
                        {"range": {specification.field: {"lt": specification.value}}}
                    ]
                }
            }
        elif isinstance(specification, LessThanEqualSpecification):
            return {
                "bool": {
                    "filter": [
                        {"range": {specification.field: {"lte": specification.value}}}
                    ]
                }
            }
        elif isinstance(specification, GreaterThanSpecification):
            return {
                "bool": {
                    "filter": [
                        {"range": {specification.field: {"gt": specification.value}}}
                    ]
                }
            }
        elif isinstance(specification, GreaterThanEqualSpecification):
            return {
                "bool": {
                    "filter": [
                        {"range": {specification.field: {"gte": specification.value}}}
                    ]
                }
            }
        raise SpecificationNotMappedToElastic(
            f"Specification '{specification}' not mapped to Elastic query."
        )



import pytest

from fractal_specifications.generic.specification import Specification


def test_dumps(equals_specification):
    assert equals_specification.dumps() == '{"op": "eq", "field": "id", "value": 1}'


def test_loads(equals_specification):
    assert (
        Specification.loads('{"op": "eq", "field": "id", "value": 1}')
        == equals_specification
    )


def test_to_dict(equals_specification):
    assert equals_specification.to_dict() == {"op": "eq", "field": "id", "value": 1}


def test_from_dict(equals_specification):
    assert (
        Specification.from_dict({"op": "eq", "field": "id", "value": 1})
        == equals_specification
    )


def test_complex_specification_serialization(complex_specification):
    assert complex_specification.to_dict() == {
        "op": "or",
        "specs": [
            {
                "op": "and",
                "specs": [
                    {"field": "id", "op": "eq", "value": 1},
                    {"field": "id", "op": "eq", "value": 1.5},
                    {"field": "id", "op": "eq", "value": False},
                    {"field": "id", "op": "eq", "value": True},
                    {"field": "id", "op": "eq", "value": None},
                    {"field": "price", "op": "gt", "value": 25},
                    {"field": "price", "op": "gte", "value": 25},
                    {"field": "price", "op": "lt", "value": 25},
                    {"field": "price", "op": "lte", "value": 25},
                    {
                        "op": "not",
                        "spec": {"field": "name", "op": "isnone", "value": None},
                    },
                    {"field": "field", "op": "contains", "value": "y"},
                ],
            },
            {
                "op": "and",
                "specs": [
                    {"field": "id", "op": "neq", "value": 1},
                    {"field": "field", "op": "in", "value": [1, 2, 3]},
                    {"field": "field", "op": "matches", "value": ".*abc.*"},
                ],
            },
            {"op": "and", "specs": [{"op": "empty"}, {"op": "empty"}]},
        ],
    }
    assert (
        Specification.from_dict(complex_specification.to_dict())
        == complex_specification
    )


def test_shorthand_ops(complex_specification):
    s = complex_specification.dumps()
    for old, new in [
        ("neq", "!="),
        ("eq", "=="),
        ("lte", "<="),
        ("lt", "<"),
        ("gte", ">="),
        ("gt", ">"),
        ("not", "!"),
        ("and", "&"),
        ("or", "|"),
    ]:
        s = s.replace(old, new)
    assert Specification.loads(s) == complex_specification


def test_complex_specification_dsl(complex_specification):
    print(complex_specification)
    assert (
        Specification.load_dsl(complex_specification.dump_dsl())
        == complex_specification
    )


def test_error_specification_dsl(complex_specification):
    from typing import Any, Collection

    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(ValueError):
        ErrorSpecification().dump_dsl()


def test_random_dsl_strings():
    for s in [
        "#",
        "# && #",
        "x == 1",
        "x == 'x'",
        "x != 'x'",
        "id != 1 && field in [1, 2, 3]",
        "a == 1 || ((b == 2) && c == 3)",
        "id == 1",
        "(id == 1)",
        "!(id == 1)",
        "(id == 1 || id == 2)",
        "id == 1 || id == 2",
        "id == 1 || id == 2 || id == 3 || id == 4",
        "(id == 1 || id == 2 || id == 3) && id == 3",
        "((id == 1 || (id == 2 || id == 3)) && id == 3)",
        "id == 1 || (id == 2 && id == 3)",
        "id == 1 || (id == 2 && id == 3 && !(name is None))",
        "!(name is None)",
        "!(name is None) && id == 1 && id == 3",
        "!(name is None) && (id == 1 || !(id == 3))",
        "!((name is None) && id == 1)",
        'field == "y"',
        'field == "y" || field == "y"',
        "field == 'y'",
        'field == "y"',
        "field contains 'y'",
        "id == 2 || field in [1, 2, 4, '2']",
        "field in [1, 2, 4, '2'] && id == 2",
        "field matches 'a*' && field in [1, 2, 4, '2']",
        'field matches "\\"',
        'field matches "\\"',
        "field matches '\\'",
        "field matches '\..*'",
        "((id == 1 && price > 25 && price >= 25 && price < 25 && price <= 25 && !(name is None) && field contains 'y' && field in [1, 2, 3] && field matches 'aaa\..*bbb'))",
        "((id == 1 && price > 25 && price >= 25 && price < 25 && price <= 25 && !(name is None) && field contains 'y') || (id != 1 && field in [1, 2, 3] && field matches '.*abc.*') || (# && #))",
        "!(id == 2)",
        "id != 2",
        "id == True",
        "id == False",
        "id == None",
        "id == -1",
        "id == 1.5",
        "field_name == 10",
        "name != 'John'",
        "age >= 18 && is_student == True",
        'roles contains "admin" || roles contains "editor"',
        "!(active == True)",
        "name in['John', 'Jane']",
        'email matches "[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}"',
        'address contains "123 Main St"',
        "#",
        "salary is None",
        "obj.id == 1",
    ]:
        spec = Specification.load_dsl(s)
        assert spec == Specification.load_dsl(spec.dump_dsl())



from typing import Any, Collection

import pytest

from fractal_specifications.contrib.google_firestore.specifications import (
    FirestoreSpecificationBuilder,
    SpecificationNotMappedToFirestore,
)


def test_build_none():
    assert FirestoreSpecificationBuilder.build(None) == None


def test_build_equals_specification(equals_specification):
    assert FirestoreSpecificationBuilder.build(equals_specification) == ("id", "==", 1)


def test_build_or_specification(or_specification):
    with pytest.raises(SpecificationNotMappedToFirestore):
        FirestoreSpecificationBuilder.build(or_specification)


def test_build_and_specification(and_specification):
    assert FirestoreSpecificationBuilder.build(and_specification) == [
        ("id", "==", 1),
        ("name", "==", "test"),
    ]


def test_build_contains_specification(contains_specification):
    assert FirestoreSpecificationBuilder.build(contains_specification) == (
        "field",
        "array-contains",
        "test",
    )


def test_build_in_specification(in_specification):
    assert FirestoreSpecificationBuilder.build(in_specification) == (
        "field",
        "in",
        [1, 2, 3],
    )


def test_build_in_empty_specification(in_empty_specification):
    assert FirestoreSpecificationBuilder.build(in_empty_specification) == None


def test_build_less_than_specification(less_than_specification):
    assert FirestoreSpecificationBuilder.build(less_than_specification) == (
        "id",
        "<",
        1,
    )


def test_build_less_than_equal_specification(less_than_equal_specification):
    assert FirestoreSpecificationBuilder.build(less_than_equal_specification) == (
        "id",
        "<=",
        1,
    )


def test_build_greater_than_specification(greater_than_specification):
    assert FirestoreSpecificationBuilder.build(greater_than_specification) == (
        "id",
        ">",
        1,
    )


def test_build_greater_than_equal_specification(greater_than_equal_specification):
    assert FirestoreSpecificationBuilder.build(greater_than_equal_specification) == (
        "id",
        ">=",
        1,
    )


def test_build_regex_string_match_specification(regex_string_match_specification):
    assert FirestoreSpecificationBuilder.build(regex_string_match_specification) == (
        "id",
        "array-contains",
        "abc",
    )


def test_build_is_none_specification(is_none_specification):
    with pytest.raises(SpecificationNotMappedToFirestore):
        FirestoreSpecificationBuilder.build(is_none_specification)


def test_build_dict_specification(dict_specification):
    assert FirestoreSpecificationBuilder.build(dict_specification) == [
        ("id", "==", 1),
        ("test", "==", 2),
    ]


def test_build_empty_specification(empty_specification):
    assert FirestoreSpecificationBuilder.build(empty_specification) == None


def test_specification_not_mapped():
    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(SpecificationNotMappedToFirestore):
        FirestoreSpecificationBuilder.build(ErrorSpecification())



from typing import Collection, Optional

from fractal_specifications.generic.collections import AndSpecification, OrSpecification
from fractal_specifications.generic.operators import (
    EqualsSpecification,
    InSpecification,
)
from fractal_specifications.generic.specification import (
    EmptySpecification,
    Specification,
)


class SpecificationNotMappedToSqlAlchemyOrm(Exception):
    pass


class SqlAlchemyOrmSpecificationBuilder:
    @staticmethod
    def build(specification: Optional[Specification] = None) -> Optional[Collection]:
        if specification is None:
            return None
        elif isinstance(specification, EmptySpecification):
            return None
        elif isinstance(specification, OrSpecification):
            return [
                s
                for spec in specification.to_collection()
                if (s := SqlAlchemyOrmSpecificationBuilder.build(spec))
            ]
        elif isinstance(specification, AndSpecification):
            return {
                k: v
                for spec in specification.to_collection()
                if (i := SqlAlchemyOrmSpecificationBuilder.build(spec))
                for k, v in dict(i).items()
                if isinstance(i, dict)
            }
        elif isinstance(specification, EqualsSpecification):
            return {specification.field: specification.value}
        elif isinstance(specification, InSpecification):
            raise SpecificationNotMappedToSqlAlchemyOrm(
                f"Specification '{specification}' not mapped to SqlAlchemy Orm query."
            )
        elif isinstance(specification.to_collection(), dict):
            return specification.to_collection()
        raise SpecificationNotMappedToSqlAlchemyOrm(
            f"Specification '{specification}' not mapped to SqlAlchemy Orm query."
        )



from typing import Any, Collection

import pytest

from fractal_specifications.contrib.elasticsearch.specifications import (
    ElasticSpecificationBuilder,
    SpecificationNotMappedToElastic,
)


def test_build_none():
    assert ElasticSpecificationBuilder.build(None) == None


def test_build_equals_specification(equals_specification):
    assert ElasticSpecificationBuilder.build(equals_specification) == {
        "match": {"id.keyword": 1}
    }


def test_build_or_specification(or_specification):
    assert ElasticSpecificationBuilder.build(or_specification) == {
        "bool": {
            "should": [
                {"match": {"id.keyword": 1}},
                {"match": {"name.keyword": "test"}},
            ]
        }
    }


def test_build_and_specification(and_specification):
    assert ElasticSpecificationBuilder.build(and_specification) == {
        "bool": {
            "must": [
                {"match": {"id.keyword": 1}},
                {"match": {"name.keyword": "test"}},
            ]
        }
    }


def test_build_contains_specification(contains_specification):
    with pytest.raises(SpecificationNotMappedToElastic):
        ElasticSpecificationBuilder.build(contains_specification)


def test_build_in_specification(in_specification):
    assert ElasticSpecificationBuilder.build(in_specification) == {
        "query_string": {"default_field": "field", "query": [1, 2, 3]}
    }


def test_build_in_empty_specification(in_empty_specification):
    assert ElasticSpecificationBuilder.build(in_empty_specification) == {
        "query_string": {"default_field": "field", "query": []}
    }


def test_build_less_than_specification(less_than_specification):
    assert ElasticSpecificationBuilder.build(less_than_specification) == {
        "bool": {"filter": [{"range": {"id": {"lt": 1}}}]}
    }


def test_build_less_than_equal_specification(less_than_equal_specification):
    assert ElasticSpecificationBuilder.build(less_than_equal_specification) == {
        "bool": {"filter": [{"range": {"id": {"lte": 1}}}]}
    }


def test_build_greater_than_specification(greater_than_specification):
    assert ElasticSpecificationBuilder.build(greater_than_specification) == {
        "bool": {"filter": [{"range": {"id": {"gt": 1}}}]}
    }


def test_build_greater_than_equal_specification(greater_than_equal_specification):
    assert ElasticSpecificationBuilder.build(greater_than_equal_specification) == {
        "bool": {"filter": [{"range": {"id": {"gte": 1}}}]}
    }


def test_build_regex_string_match_specification(regex_string_match_specification):
    with pytest.raises(SpecificationNotMappedToElastic):
        assert ElasticSpecificationBuilder.build(regex_string_match_specification) == {}


def test_build_is_none_specification(is_none_specification):
    with pytest.raises(SpecificationNotMappedToElastic):
        ElasticSpecificationBuilder.build(is_none_specification)


def test_build_dict_specification(dict_specification):
    with pytest.raises(SpecificationNotMappedToElastic):
        ElasticSpecificationBuilder.build(dict_specification)


def test_build_empty_specification(empty_specification):
    assert ElasticSpecificationBuilder.build(empty_specification) == None


def test_specification_not_mapped():
    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(SpecificationNotMappedToElastic):
        ElasticSpecificationBuilder.build(ErrorSpecification())



import pytest


@pytest.fixture
def equals_specification():
    from fractal_specifications.generic.operators import EqualsSpecification

    return EqualsSpecification("id", 1)


@pytest.fixture
def other_equals_specification():
    from fractal_specifications.generic.operators import EqualsSpecification

    return EqualsSpecification("name", "test")


@pytest.fixture
def or_specification(
    empty_specification, equals_specification, other_equals_specification
):
    from fractal_specifications.generic.collections import OrSpecification

    return OrSpecification(
        [empty_specification, equals_specification, other_equals_specification]
    )


@pytest.fixture
def and_specification(
    empty_specification, equals_specification, other_equals_specification
):
    from fractal_specifications.generic.collections import AndSpecification

    return AndSpecification(
        [empty_specification, equals_specification, other_equals_specification]
    )


@pytest.fixture
def contains_specification():
    from fractal_specifications.generic.operators import ContainsSpecification

    return ContainsSpecification("field", "test")


@pytest.fixture
def in_specification():
    from fractal_specifications.generic.operators import InSpecification

    return InSpecification("field", [1, 2, 3])


@pytest.fixture
def in_empty_specification():
    from fractal_specifications.generic.operators import InSpecification

    return InSpecification("field", [])


@pytest.fixture
def less_than_specification():
    from fractal_specifications.generic.operators import LessThanSpecification

    return LessThanSpecification("id", 1)


@pytest.fixture
def less_than_equal_specification():
    from fractal_specifications.generic.operators import LessThanEqualSpecification

    return LessThanEqualSpecification("id", 1)


@pytest.fixture
def greater_than_specification():
    from fractal_specifications.generic.operators import GreaterThanSpecification

    return GreaterThanSpecification("id", 1)


@pytest.fixture
def greater_than_equal_specification():
    from fractal_specifications.generic.operators import GreaterThanEqualSpecification

    return GreaterThanEqualSpecification("id", 1)


@pytest.fixture
def regex_string_match_specification():
    from fractal_specifications.generic.operators import RegexStringMatchSpecification

    return RegexStringMatchSpecification("id", "abc")


@pytest.fixture
def is_none_specification():
    from fractal_specifications.generic.operators import IsNoneSpecification

    return IsNoneSpecification("field")


@pytest.fixture
def dict_specification():
    from fractal_specifications.generic.specification import Specification

    class DictSpecification(Specification):
        def __init__(self, collection):
            self.collection = collection

        def is_satisfied_by(self, obj) -> bool:
            raise NotImplementedError

        def to_collection(self) -> dict:
            return self.collection

        def __str__(self):
            return f"{self.__class__.__name__}({self.collection})"

    return DictSpecification({"id": 1, "test": 2})


@pytest.fixture
def empty_specification():
    from fractal_specifications.generic.specification import EmptySpecification

    return EmptySpecification()


@pytest.fixture
def complex_specification():
    from fractal_specifications.generic import operators
    from fractal_specifications.generic.specification import EmptySpecification

    return (
        operators.EqualsSpecification("id", 1)
        & operators.EqualsSpecification("id", 1.5)
        & operators.EqualsSpecification("id", False)
        & operators.EqualsSpecification("id", True)
        & operators.EqualsSpecification("id", None)
        & operators.GreaterThanSpecification("price", 25)
        & operators.GreaterThanEqualSpecification("price", 25)
        & operators.LessThanSpecification("price", 25)
        & operators.LessThanEqualSpecification("price", 25)
        & operators.NotSpecification(operators.IsNoneSpecification("name"))
        & operators.ContainsSpecification("field", "y")
        | operators.NotEqualsSpecification("id", 1)
        & operators.InSpecification("field", [1, 2, 3])
        & operators.RegexStringMatchSpecification("field", ".*abc.*")
        | (EmptySpecification() & EmptySpecification())
    )



from typing import Any, Collection

import pytest

from fractal_specifications.contrib.mongo.specifications import (
    MongoSpecificationBuilder,
    SpecificationNotMappedToMongo,
)


def test_build_none():
    assert MongoSpecificationBuilder.build(None) == None


def test_build_equals_specification(equals_specification):
    assert MongoSpecificationBuilder.build(equals_specification) == {"id": {"$eq": 1}}


def test_build_or_specification(or_specification):
    assert MongoSpecificationBuilder.build(or_specification) == {
        "$or": [{"id": {"$eq": 1}}, {"name": {"$eq": "test"}}]
    }


def test_build_and_specification(and_specification):
    assert MongoSpecificationBuilder.build(and_specification) == {
        "$and": [{"id": {"$eq": 1}}, {"name": {"$eq": "test"}}]
    }


def test_build_contains_specification(contains_specification):
    with pytest.raises(SpecificationNotMappedToMongo):
        MongoSpecificationBuilder.build(contains_specification)


def test_build_in_specification(in_specification):
    assert MongoSpecificationBuilder.build(in_specification) == {
        "field": {"$in": [1, 2, 3]}
    }


def test_build_in_empty_specification(in_empty_specification):
    assert MongoSpecificationBuilder.build(in_empty_specification) == {
        "field": {"$in": []}
    }


def test_build_less_than_specification(less_than_specification):
    assert MongoSpecificationBuilder.build(less_than_specification) == {
        "id": {"$lt": 1}
    }


def test_build_less_than_equal_specification(less_than_equal_specification):
    assert MongoSpecificationBuilder.build(less_than_equal_specification) == {
        "id": {"$lte": 1}
    }


def test_build_greater_than_specification(greater_than_specification):
    assert MongoSpecificationBuilder.build(greater_than_specification) == {
        "id": {"$gt": 1}
    }


def test_build_greater_than_equal_specification(greater_than_equal_specification):
    assert MongoSpecificationBuilder.build(greater_than_equal_specification) == {
        "id": {"$gte": 1}
    }


def test_build_regex_string_match_specification(regex_string_match_specification):
    assert MongoSpecificationBuilder.build(regex_string_match_specification) == {
        "id": {"$regex": ".*abc.*"}
    }


def test_build_is_none_specification(is_none_specification):
    with pytest.raises(SpecificationNotMappedToMongo):
        MongoSpecificationBuilder.build(is_none_specification)


def test_build_dict_specification(dict_specification):
    assert MongoSpecificationBuilder.build(dict_specification) == {
        "$and": [{"id": {"$eq": 1}}, {"test": {"$eq": 2}}]
    }


def test_build_empty_specification(empty_specification):
    assert MongoSpecificationBuilder.build(empty_specification) == None
    assert (
        MongoSpecificationBuilder.build(empty_specification & empty_specification)
        == None
    )
    assert (
        MongoSpecificationBuilder.build(empty_specification | empty_specification)
        == None
    )


def test_specification_not_mapped():
    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(SpecificationNotMappedToMongo):
        MongoSpecificationBuilder.build(ErrorSpecification())



from functools import reduce
from typing import Callable, Dict, Iterator, Optional, Type

import pandas as pd  # type: ignore

from fractal_specifications.generic.specification import (
    EmptySpecification,
    Specification,
)


class SpecificationNotMappedToPandas(Exception):
    pass


class PandasSpecificationBuilder:
    @classmethod
    def build(
        cls,
        specification: Optional[Specification] = None,
        *,
        return_mask=False,
    ) -> Optional[Callable[[pd.DataFrame], pd.Series]]:
        if specification is None or isinstance(specification, EmptySpecification):
            return None
        if builder := cls._spec_builders().get(type(specification)):
            if f := builder(specification):
                return lambda df: f(df) if return_mask else df[f(df)]
        raise SpecificationNotMappedToPandas(
            f"Specification '{specification}' not mapped to Pandas query."
        )

    @classmethod
    def _spec_builders(
        cls,
    ) -> Dict[Type[Specification], Callable]:
        from fractal_specifications.generic import collections, operators

        return {
            collections.AndSpecification: lambda s: reduce(
                lambda x, y: lambda df: x(df) & y(df), cls._build_collection(s)
            ),
            collections.OrSpecification: lambda s: reduce(
                lambda x, y: lambda df: x(df) | y(df), cls._build_collection(s)
            ),
            operators.EqualsSpecification: lambda s: lambda df: df[s.field] == s.value,
            operators.InSpecification: lambda s: lambda df: df[s.field].isin(s.value),
            operators.LessThanSpecification: lambda s: lambda df: df[s.field] < s.value,
            operators.LessThanEqualSpecification: lambda s: lambda df: df[s.field]
            <= s.value,
            operators.GreaterThanSpecification: lambda s: lambda df: df[s.field]
            > s.value,
            operators.GreaterThanEqualSpecification: lambda s: lambda df: df[s.field]
            >= s.value,
            operators.IsNoneSpecification: lambda s: lambda df: df[s.field].isna(),
        }

    @classmethod
    def _build_collection(
        cls, specification
    ) -> Iterator[Callable[[pd.DataFrame], pd.Series]]:
        for spec in specification.to_collection():
            if s := cls.build(spec, return_mask=True):
                yield s


class PandasIndexSpecificationBuilder(PandasSpecificationBuilder):
    @classmethod
    def _spec_builders(
        cls,
    ) -> Dict[Type[Specification], Callable]:
        from fractal_specifications.generic import collections, operators

        return {
            collections.AndSpecification: lambda s: reduce(
                lambda x, y: lambda df: x(df) & y(df), cls._build_collection(s)
            ),
            collections.OrSpecification: lambda s: reduce(
                lambda x, y: lambda df: x(df) | y(df), cls._build_collection(s)
            ),
            operators.EqualsSpecification: lambda s: lambda df: df.index.get_level_values(
                s.field
            )
            == s.value,
            operators.InSpecification: lambda s: lambda df: df.index.get_level_values(
                s.field
            ).isin(s.value),
            operators.LessThanSpecification: lambda s: lambda df: df.index.get_level_values(
                s.field
            )
            < s.value,
            operators.LessThanEqualSpecification: lambda s: lambda df: df.index.get_level_values(
                s.field
            )
            <= s.value,
            operators.GreaterThanSpecification: lambda s: lambda df: df.index.get_level_values(
                s.field
            )
            > s.value,
            operators.GreaterThanEqualSpecification: lambda s: lambda df: df.index.get_level_values(
                s.field
            )
            >= s.value,
            operators.IsNoneSpecification: lambda s: lambda df: df.index.get_level_values(
                s.field
            ).isna(),
        }



from lark import Transformer

from fractal_specifications.generic.collections import AndSpecification, OrSpecification
from fractal_specifications.generic.operators import (
    ContainsSpecification,
    EqualsSpecification,
    GreaterThanEqualSpecification,
    GreaterThanSpecification,
    InSpecification,
    IsNoneSpecification,
    LessThanEqualSpecification,
    LessThanSpecification,
    NotEqualsSpecification,
    NotSpecification,
    RegexStringMatchSpecification,
)
from fractal_specifications.generic.specification import EmptySpecification

grammar = r"""
    ?start: expression
    expression: and_expression | or_expression | comparison_expression
    or_expression: expression ("||" comparison_expression)+
    and_expression: expression ("&&" comparison_expression)+
    comparison_expression: field_name comparison_operator field_value
        | atom_expression -> atom_expression
        | not_expression -> atom_expression
        | field_name "in" "[" field_values "]" -> in_expression
        | field_name "matches" string_value -> match_expression
        | field_name "is" "None" -> is_none_expression
        | field_name "contains" field_value -> contains_expression
        | empty_expression
    empty_expression: "#"
    not_expression: "!" atom_expression
    atom_expression: "(" expression ")"
    field_values: (field_value ",")* field_value?
    field_value: string_value | number_value | boolean_value | none_value
    comparison_operator: "==" -> eq_op
        | "!=" -> neq_op
        | "<" -> lt_op
        | "<=" -> lte_op
        | ">" -> gt_op
        | ">="-> gte_op
    field_name: CNAME ("." CNAME)*
    string_value: ESCAPED_STRING | DOUBLE_QUOTED_STRING | SINGLE_QUOTED_STRING
    number_value: SIGNED_FLOAT | SIGNED_INT
    boolean_value: "True" -> true | "False" -> false
    none_value: "None" -> none
    DOUBLE_QUOTED_STRING  : /"[^"]*"/
    SINGLE_QUOTED_STRING  : /'[^']*'/
    %import common.ESCAPED_STRING
    %import common.SIGNED_FLOAT
    %import common.SIGNED_INT
    %import common.CNAME
    %import common.WS
    %ignore WS
"""


class DSLTransformer(Transformer):
    def expression(self, items):
        return items[0]

    def and_expression(self, items):
        return AndSpecification(items)

    def or_expression(self, items):
        return OrSpecification(items)

    def not_expression(self, items):
        return NotSpecification(items[0])

    def comparison_expression(self, items):
        from fractal_specifications.generic.specification import Specification

        if issubclass(type(items[0]), Specification):
            return items[0]
        field_name, op, value = items
        if op == "==":
            return EqualsSpecification(field_name, value)
        elif op == "!=":
            return NotEqualsSpecification(field_name, value)
        elif op == "<":
            return LessThanSpecification(field_name, value)
        elif op == "<=":
            return LessThanEqualSpecification(field_name, value)
        elif op == ">":
            return GreaterThanSpecification(field_name, value)
        elif op == ">=":
            return GreaterThanEqualSpecification(field_name, value)

    def empty_expression(self, items):
        return EmptySpecification()

    def atom_expression(self, items):
        return items[0]

    def eq_op(self, items):
        return "=="

    def neq_op(self, items):
        return "!="

    def lt_op(self, items):
        return "<"

    def lte_op(self, items):
        return "<="

    def gt_op(self, items):
        return ">"

    def gte_op(self, items):
        return ">="

    def in_expression(self, items):
        field_name, values = items
        return InSpecification(field_name, values)

    def match_expression(self, items):
        field_name, value = items
        return RegexStringMatchSpecification(field_name, value)

    def contains_expression(self, items):
        field_name, value = items
        return ContainsSpecification(field_name, value)

    def field_name(self, items):
        return ".".join(items)

    def is_none_expression(self, items):
        return IsNoneSpecification(items[0])

    def field_values(self, tokens):
        return tokens

    def field_value(self, tokens):
        token = tokens[0]
        if token is None:
            return None
        elif type(token) is str:
            return token
        elif type(token) is bool:
            return token
        elif token.type == "SIGNED_INT":
            return int(token)
        elif token.type == "SIGNED_FLOAT":
            return float(token)

    def string_value(self, items):
        return str(items[0])[1:-1]

    def number_value(self, items):
        return items[0]

    def true(self, items):
        return True

    def false(self, items):
        return False

    def none(self, items):
        return None



from dataclasses import make_dataclass
from typing import List

from fractal_specifications.generic.operators import (
    ContainsSpecification,
    EqualsSpecification,
    FieldValueSpecification,
    GreaterThanEqualSpecification,
    GreaterThanSpecification,
    InSpecification,
    IsNoneSpecification,
    LessThanEqualSpecification,
    LessThanSpecification,
    NotEqualsSpecification,
    NotSpecification,
    RegexStringMatchSpecification,
)


def test_not_specification():
    spec = NotSpecification(EqualsSpecification("id", 2))
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_not_specification_to_collection():
    spec = NotSpecification(EqualsSpecification("id", 2))
    assert spec.to_collection() == [EqualsSpecification("id", 2)]


def test_not_specification_str():
    spec = NotSpecification(EqualsSpecification("id", 2))
    assert spec.__str__() == "NotSpecification(EqualsSpecification(id=2))"


def test_not_specification_eq():
    assert NotSpecification(EqualsSpecification("id", 2)) == NotSpecification(
        EqualsSpecification("id", 2)
    )


def test_field_value_specification():
    spec = FieldValueSpecification("field", "value")
    assert spec.to_collection() == {"field", "value"}


def test_in_specification():
    spec = InSpecification("id", [1, 2, 3])
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_equals_specification():
    spec = EqualsSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_equals_specification_dot():
    spec = EqualsSpecification("obj.id", 1)
    DC1 = make_dataclass("DC1", [("id", int)])
    DC2 = make_dataclass("DC2", [("obj", DC1)])
    assert spec.is_satisfied_by(DC2(DC1(**dict(id=1))))


def test_not_equals_specification():
    spec = NotEqualsSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=2)))


def test_less_than_specification():
    spec = LessThanSpecification("id", 2)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_less_than_equal_specification():
    spec = LessThanEqualSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_greater_than_specification():
    spec = GreaterThanSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=2)))


def test_greater_than_equal_specification():
    spec = GreaterThanEqualSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_contains_specification():
    spec = ContainsSpecification("name", "a")
    DC = make_dataclass("DC", [("name", str)])
    assert spec.is_satisfied_by(DC(**dict(name="fractal")))


def test_contains_dot_specification():
    spec = ContainsSpecification("obj.name", "a")
    DC = make_dataclass("DC", [("name", str)])
    OBJ = make_dataclass("OBJ", [("obj", DC)])
    assert spec.is_satisfied_by(OBJ(obj=DC(name="fractal")))


def test_contains_double_underscore_specification():
    spec = ContainsSpecification("obj__name", "a")
    DC = make_dataclass("DC", [("name", str)])
    OBJ = make_dataclass("OBJ", [("obj", DC)])
    assert spec.is_satisfied_by(OBJ(obj=DC(name="fractal")))


def test_contains_none_specification():
    spec = ContainsSpecification("name", "a")
    DC = make_dataclass("DC", [("name", None)])
    assert not spec.is_satisfied_by(DC(**dict(name=None)))


def test_list_contains_specification():
    spec = ContainsSpecification("names", "fractal")
    DC = make_dataclass("DC", [("names", list)])
    assert spec.is_satisfied_by(DC(**dict(names=["fractal"])))


def test_regex_string_match_specification():
    spec = RegexStringMatchSpecification("name", "^f.*l$")
    DC = make_dataclass("DC", [("name", str)])
    assert spec.is_satisfied_by(DC(**dict(name="fractal")))


def test_is_none_specification():
    spec = IsNoneSpecification("name")
    DC = make_dataclass("DC", [("name", str)])
    assert spec.is_satisfied_by(DC(**dict(name=None)))


def test_is_none_specification_str():
    spec = IsNoneSpecification("name")
    assert spec.__str__() == "IsNoneSpecification(name)"


def test_is_none_specification_repr():
    spec = IsNoneSpecification("name")
    assert spec.__repr__() == "IsNoneSpecification(name)"


def test_pre_process_equals_specification():
    spec = EqualsSpecification("name", "abc", lambda i: i.lower())
    DC = make_dataclass("DC", [("name", str)])
    assert spec.is_satisfied_by(DC(**dict(name="ABC")))


def test_pre_process_contains_specification():
    spec = ContainsSpecification("name", "A", lambda i: i.upper())
    DC = make_dataclass("DC", [("name", str)])
    assert spec.is_satisfied_by(DC(**dict(name="fractal")))


def test_pre_process_list_contains_specification():
    spec = ContainsSpecification("roles", "ADMIN", lambda i: [r.upper() for r in i])
    DC = make_dataclass("DC", [("roles", List[str])])
    assert spec.is_satisfied_by(DC(**dict(roles=["admin", "owner"])))


def test_pre_process_regex_string_match_specification():
    spec = RegexStringMatchSpecification(
        "roles", "a.*n", lambda i: ",".join([r.lower() for r in i])
    )
    DC = make_dataclass("DC", [("roles", List[str])])
    assert spec.is_satisfied_by(DC(**dict(roles=["admin", "owner"])))



from typing import Any, Collection

import pytest
from django.db.models import Q  # type: ignore

from fractal_specifications.contrib.django.specifications import (
    DjangoOrmSpecificationBuilder,
    SpecificationNotMappedToDjangoOrm,
)


def test_build_none():
    assert DjangoOrmSpecificationBuilder.build(None) == None


def test_build_equals_specification(equals_specification):
    assert DjangoOrmSpecificationBuilder.build(equals_specification) == Q(id=1)


def test_build_or_specification(or_specification):
    assert DjangoOrmSpecificationBuilder.build(or_specification) == Q(id=1) | Q(
        name="test"
    )


def test_build_and_specification(and_specification):
    assert DjangoOrmSpecificationBuilder.build(and_specification) == Q(
        id=1, name="test"
    )


def test_build_contains_specification(contains_specification):
    with pytest.raises(SpecificationNotMappedToDjangoOrm):
        DjangoOrmSpecificationBuilder.build(contains_specification)


def test_build_in_specification(in_specification):
    assert DjangoOrmSpecificationBuilder.build(in_specification) == Q(
        field__in=[1, 2, 3]
    )


def test_build_in_empty_specification(in_empty_specification):
    assert DjangoOrmSpecificationBuilder.build(in_empty_specification) == Q(
        field__in=[]
    )


def test_build_less_than_specification(less_than_specification):
    assert DjangoOrmSpecificationBuilder.build(less_than_specification) == Q(id__lt=1)


def test_build_less_than_equal_specification(less_than_equal_specification):
    assert DjangoOrmSpecificationBuilder.build(less_than_equal_specification) == Q(
        id__lte=1
    )


def test_build_greater_than_specification(greater_than_specification):
    assert DjangoOrmSpecificationBuilder.build(greater_than_specification) == Q(
        id__gt=1
    )


def test_build_greater_than_equal_specification(greater_than_equal_specification):
    assert DjangoOrmSpecificationBuilder.build(greater_than_equal_specification) == Q(
        id__gte=1
    )


def test_build_regex_string_match_specification(regex_string_match_specification):
    assert DjangoOrmSpecificationBuilder.build(regex_string_match_specification) == Q(
        id__regex=".*abc.*"
    )


def test_build_is_none_specification(is_none_specification):
    assert DjangoOrmSpecificationBuilder.build(is_none_specification) == Q(
        field__isnull=True
    )


def test_build_dict_specification(dict_specification):
    assert DjangoOrmSpecificationBuilder.build(dict_specification) == Q(id=1, test=2)


def test_build_empty_specification(empty_specification):
    assert DjangoOrmSpecificationBuilder.build(empty_specification) == None


def test_create_q_dict():
    assert isinstance(DjangoOrmSpecificationBuilder._create_q({"id": 1}), Q)
    assert DjangoOrmSpecificationBuilder._create_q({"id": 1}).children == [("id", 1)]


def test_create_q_list():
    assert isinstance(DjangoOrmSpecificationBuilder._create_q([("id", 1)]), Q)
    assert DjangoOrmSpecificationBuilder._create_q([("id", 1)]).children == [("id", 1)]


def test_specification_not_mapped():
    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(SpecificationNotMappedToDjangoOrm):
        DjangoOrmSpecificationBuilder.build(ErrorSpecification())



from typing import Any, Collection

import pandas as pd  # type: ignore
import pytest  # type: ignore

from fractal_specifications.contrib.pandas.specifications import (
    PandasIndexSpecificationBuilder,
    PandasSpecificationBuilder,
    SpecificationNotMappedToPandas,
)

df = pd.DataFrame(
    {
        "id": [1, 2, 3, 4],
        "name": ["test", "test", "test", "test"],
        "field": [3, 4, 5, 6],
    }
)

dfi = df.set_index(["id", "name", "field"])


def test_build_none():
    assert PandasSpecificationBuilder.build(None) == None


def test_build_equals_specification(equals_specification):
    assert (
        PandasSpecificationBuilder.build(equals_specification)(df).to_dict()
        == df[df.id == 1].to_dict()
    )


def test_build_or_specification(or_specification):
    assert (
        PandasSpecificationBuilder.build(or_specification)(df).to_dict()
        == (df[(df.id == 1) | (df.name == "test")]).to_dict()
    )


def test_build_and_specification(and_specification):
    assert (
        PandasSpecificationBuilder.build(and_specification)(df).to_dict()
        == (df[(df.id == 1) & (df.name == "test")]).to_dict()
    )


def test_build_contains_specification(contains_specification):
    with pytest.raises(SpecificationNotMappedToPandas):
        PandasSpecificationBuilder.build(contains_specification)


def test_build_in_specification(in_specification):
    assert (
        PandasSpecificationBuilder.build(in_specification)(df).to_dict()
        == df[df.field.isin([1, 2, 3])].to_dict()
    )


def test_build_in_empty_specification(in_empty_specification):
    assert (
        PandasSpecificationBuilder.build(in_empty_specification)(df).to_dict()
        == df[df.field.isin([])].to_dict()
    )


def test_build_less_than_specification(less_than_specification):
    assert (
        PandasSpecificationBuilder.build(less_than_specification)(df).to_dict()
        == df[df.id < 1].to_dict()
    )


def test_build_less_than_equal_specification(less_than_equal_specification):
    assert (
        PandasSpecificationBuilder.build(less_than_equal_specification)(df).to_dict()
        == df[df.id <= 1].to_dict()
    )


def test_build_greater_than_specification(greater_than_specification):
    assert (
        PandasSpecificationBuilder.build(greater_than_specification)(df).to_dict()
        == df[df.id > 1].to_dict()
    )


def test_build_greater_than_equal_specification(greater_than_equal_specification):
    assert (
        PandasSpecificationBuilder.build(greater_than_equal_specification)(df).to_dict()
        == df[df.id >= 1].to_dict()
    )


def test_build_regex_string_match_specification(regex_string_match_specification):
    with pytest.raises(SpecificationNotMappedToPandas):
        PandasSpecificationBuilder.build(regex_string_match_specification)


def test_build_is_none_specification(is_none_specification):
    assert PandasSpecificationBuilder.build(is_none_specification)(df).to_dict() == {
        "field": {},
        "id": {},
        "name": {},
    }


def test_build_dict_specification(dict_specification):
    with pytest.raises(SpecificationNotMappedToPandas):
        PandasSpecificationBuilder.build(dict_specification)


def test_build_empty_specification(empty_specification):
    assert PandasSpecificationBuilder.build(empty_specification) == None


def test_specification_not_mapped():
    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(SpecificationNotMappedToPandas):
        PandasSpecificationBuilder.build(ErrorSpecification())


def test_build_with_index_none():
    assert PandasIndexSpecificationBuilder.build(None) == None


def test_build_with_index_equals_specification(equals_specification):
    assert (
        PandasIndexSpecificationBuilder.build(equals_specification)(dfi).to_dict()
        == dfi[dfi.index.get_level_values("id") == 1].to_dict()
    )


def test_build_with_index_or_specification(or_specification):
    assert (
        PandasIndexSpecificationBuilder.build(or_specification)(dfi).to_dict()
        == (
            dfi[
                (dfi.index.get_level_values("id") == 1)
                | (dfi.index.get_level_values("name") == "test")
            ]
        ).to_dict()
    )


def test_build_with_index_and_specification(and_specification):
    assert (
        PandasIndexSpecificationBuilder.build(and_specification)(dfi).to_dict()
        == (
            dfi[
                (dfi.index.get_level_values("id") == 1)
                & (dfi.index.get_level_values("name") == "test")
            ]
        ).to_dict()
    )


def test_build_with_index_contains_specification(contains_specification):
    with pytest.raises(SpecificationNotMappedToPandas):
        PandasIndexSpecificationBuilder.build(contains_specification)


def test_build_with_index_in_specification(in_specification):
    assert (
        PandasIndexSpecificationBuilder.build(in_specification)(dfi).to_dict()
        == dfi[dfi.index.get_level_values("field").isin([1, 2, 3])].to_dict()
    )


def test_build_with_index_in_empty_specification(in_empty_specification):
    assert (
        PandasIndexSpecificationBuilder.build(in_empty_specification)(dfi).to_dict()
        == dfi[dfi.index.get_level_values("field").isin([])].to_dict()
    )


def test_build_with_index_less_than_specification(less_than_specification):
    assert (
        PandasIndexSpecificationBuilder.build(less_than_specification)(dfi).to_dict()
        == dfi[dfi.index.get_level_values("id") < 1].to_dict()
    )


def test_build_with_index_less_than_equal_specification(less_than_equal_specification):
    assert (
        PandasIndexSpecificationBuilder.build(less_than_equal_specification)(
            dfi
        ).to_dict()
        == dfi[dfi.index.get_level_values("id") <= 1].to_dict()
    )


def test_build_with_index_greater_than_specification(greater_than_specification):
    assert (
        PandasIndexSpecificationBuilder.build(greater_than_specification)(dfi).to_dict()
        == dfi[dfi.index.get_level_values("id") > 1].to_dict()
    )


def test_build_with_index_greater_than_equal_specification(
    greater_than_equal_specification,
):
    assert (
        PandasIndexSpecificationBuilder.build(greater_than_equal_specification)(
            dfi
        ).to_dict()
        == dfi[dfi.index.get_level_values("id") >= 1].to_dict()
    )


def test_build_with_index_regex_string_match_specification(
    regex_string_match_specification,
):
    with pytest.raises(SpecificationNotMappedToPandas):
        PandasIndexSpecificationBuilder.build(regex_string_match_specification)


def test_build_with_index_is_none_specification(is_none_specification):
    assert (
        PandasIndexSpecificationBuilder.build(is_none_specification)(dfi).to_dict()
        == {}
    )


def test_build_with_index_dict_specification(dict_specification):
    with pytest.raises(SpecificationNotMappedToPandas):
        PandasIndexSpecificationBuilder.build(dict_specification)


def test_build_with_index_empty_specification(empty_specification):
    assert PandasIndexSpecificationBuilder.build(empty_specification) == None


def test_specification_not_mapped_with_index():
    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(SpecificationNotMappedToPandas):
        PandasIndexSpecificationBuilder.build(ErrorSpecification())



from typing import Any, Callable, Collection, List

from fractal_specifications.generic.specification import Specification


class NotSpecification(Specification):
    def __init__(self, specification: Specification):
        self.specification = specification

    def is_satisfied_by(self, obj: Any) -> bool:
        return not self.specification.is_satisfied_by(obj)

    def to_collection(self) -> Collection:
        return [self.specification]

    def __str__(self):
        return f"{self.__class__.__name__}({self.specification})"

    def __eq__(self, other):
        return (
            isinstance(other, NotSpecification)
            and self.specification == other.specification
        )

    def __hash__(self):
        return hash(self.specification)

    def to_dict(self):
        return {
            "op": self.name(),
            "spec": self.specification.to_dict(),
        }

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(specification=Specification.from_dict(d["spec"]))


class FieldValueSpecification(Specification):
    def __init__(self, field: str, value: Any, pre_processor: Callable = lambda i: i):
        self.field = field
        self.value = value
        self.pre_processor = pre_processor

    def __str__(self):
        return f"{self.__class__.__name__}({self.field}={self.value})"

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.field == other.field
            and self.value == other.value
        )

    def __hash__(self):
        return hash((self.field, self.value))

    def is_satisfied_by(self, obj: Any) -> bool:
        raise NotImplementedError

    def to_collection(self) -> Collection:
        return {self.field, self.value}


def _get_value(obj: Any, field: str) -> Any:
    lookup_separator = "__" if "__" in field else "."
    for f in field.split(lookup_separator):
        obj = getattr(obj, f)
    return obj


class InSpecification(FieldValueSpecification):
    def __init__(self, field: str, values: List[Any]):
        super(InSpecification, self).__init__(field, values)

    def __hash__(self):
        return hash((self.field, tuple(self.value)))

    def is_satisfied_by(self, obj: Any) -> bool:
        return self.pre_processor(_get_value(obj, self.field)) in self.value

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(field=d["field"], values=d["value"])


class EqualsSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return self.pre_processor(_get_value(obj, self.field)) == self.value

    @classmethod
    def name(cls):
        return "eq"


class NotEqualsSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return self.pre_processor(_get_value(obj, self.field)) != self.value

    @classmethod
    def name(cls):
        return "neq"


class LessThanSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return self.pre_processor(_get_value(obj, self.field)) < self.value

    @classmethod
    def name(cls):
        return "lt"


class LessThanEqualSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return self.pre_processor(_get_value(obj, self.field)) <= self.value

    @classmethod
    def name(cls):
        return "lte"


class GreaterThanSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return self.pre_processor(_get_value(obj, self.field)) > self.value

    @classmethod
    def name(cls):
        return "gt"


class GreaterThanEqualSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return self.pre_processor(_get_value(obj, self.field)) >= self.value

    @classmethod
    def name(cls):
        return "gte"


class ContainsSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        value = self.pre_processor(_get_value(obj, self.field))
        if not value:
            return False
        return self.value in value


class RegexStringMatchSpecification(ContainsSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        import re

        return bool(
            re.match(self.value, self.pre_processor(_get_value(obj, self.field)))
        )

    @classmethod
    def name(cls):
        return "matches"


class IsNoneSpecification(FieldValueSpecification):
    def __init__(self, field: str):
        super(IsNoneSpecification, self).__init__(field, None)

    def __str__(self):
        return f"{self.__class__.__name__}({self.field})"

    def is_satisfied_by(self, obj: Any) -> bool:
        return self.pre_processor(_get_value(obj, self.field)) is None

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(field=d["field"])



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from functools import lru_cache
from typing import Any, Collection, Iterator, Optional, Type, TypeVar


@lru_cache
def all_specifications():
    def get_subclasses(spec):
        for sub in spec.__subclasses__():
            yield sub
            for subsub in get_subclasses(sub):
                yield subsub

    from fractal_specifications.generic import collections, operators

    return {
        **{spec.name(): spec for spec in get_subclasses(Specification)},
        **{
            "==": operators.EqualsSpecification,
            "!=": operators.NotEqualsSpecification,
            "<": operators.LessThanSpecification,
            "<=": operators.LessThanEqualSpecification,
            ">": operators.GreaterThanSpecification,
            ">=": operators.GreaterThanEqualSpecification,
            "!": operators.NotSpecification,
            "&": collections.AndSpecification,
            "|": collections.OrSpecification,
        },
    }


def _parse_specification_item(
    field_op: str, value: Any, lookup_separator: str
) -> Optional[Specification]:
    parts = field_op.split("__")
    field = lookup_separator.join(parts[:-1])
    op = parts[-1]
    if spec := all_specifications().get(op, None):
        return spec(field, value)
    return all_specifications()["=="](lookup_separator.join(parts), value)


def parse_specification(lookup_separator: str, **kwargs) -> Iterator[Specification]:
    for field_op, value in kwargs.items():
        if spec := _parse_specification_item(field_op, value, lookup_separator):
            yield spec


SpecificationSubType = TypeVar("SpecificationSubType", bound="Specification")


class Specification(ABC):
    @abstractmethod
    def is_satisfied_by(self, obj: Any) -> bool:
        raise NotImplementedError

    @abstractmethod
    def to_collection(self) -> Collection:
        raise NotImplementedError

    def And(self, specification: "Specification") -> "Specification":
        from fractal_specifications.generic.collections import AndSpecification

        if isinstance(specification, AndSpecification):
            return specification.And(self)
        return AndSpecification([self, specification])

    def Or(self, specification: "Specification") -> "Specification":
        from fractal_specifications.generic.collections import OrSpecification

        if isinstance(specification, OrSpecification):
            return specification.Or(self)
        return OrSpecification([self, specification])

    def __and__(self, other):
        return self.And(other)

    def __or__(self, other):
        return self.Or(other)

    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def Not(specification: "Specification") -> "Specification":
        from fractal_specifications.generic.operators import NotSpecification

        return NotSpecification(specification)

    @staticmethod
    def parse(_lookup_separator=".", **kwargs):
        specs = list(parse_specification(lookup_separator=_lookup_separator, **kwargs))
        if len(specs) > 1:
            from fractal_specifications.generic.collections import AndSpecification

            return AndSpecification(specs)
        elif len(specs) == 1:
            return specs[0]
        return None

    def to_dict(self):
        return {
            **{
                "op": self.name(),
            },
            **{
                key: value
                for key, value in self.__dict__.items()
                if not callable(value)
            },
        }

    @classmethod
    def from_dict(cls, d: dict):
        name = d.pop("op")
        return all_specifications()[name]._from_dict(d)

    @classmethod
    def _from_dict(cls: Type[SpecificationSubType], d: dict):
        return cls(**d)

    def dumps(self) -> str:
        return json.dumps(self.to_dict())

    @staticmethod
    def loads(s: str) -> Specification:
        return Specification.from_dict(json.loads(s))

    @classmethod
    def name(cls) -> str:
        return cls.__name__[:-13].lower()  # -Specification

    def dump_dsl(self) -> Optional[str]:
        from fractal_specifications.generic import collections, operators

        if isinstance(self, operators.NotSpecification):
            child = self.specification.dump_dsl()
            return f"!({child})"
        elif isinstance(self, operators.FieldValueSpecification):
            lhs = self.field
            operator = {
                operators.EqualsSpecification.__name__: "==",
                operators.NotEqualsSpecification.__name__: "!=",
                operators.GreaterThanSpecification.__name__: ">",
                operators.GreaterThanEqualSpecification.__name__: ">=",
                operators.LessThanSpecification.__name__: "<",
                operators.LessThanEqualSpecification.__name__: "<=",
                operators.InSpecification.__name__: "in",
                operators.ContainsSpecification.__name__: "contains",
                operators.IsNoneSpecification.__name__: "is None",
                operators.RegexStringMatchSpecification.__name__: "matches",
            }[self.__class__.__name__]
            if isinstance(self, operators.IsNoneSpecification):
                return f"{lhs} {operator}"
            rhs = f'"{self.value}"' if type(self.value) is str else repr(self.value)
            return f"{lhs} {operator} {rhs}"
        elif isinstance(
            self, (collections.AndSpecification, collections.OrSpecification)
        ):
            op = {
                collections.AndSpecification: " && ",
                collections.OrSpecification: " || ",
            }[type(self)]
            child_strings = [
                val for child in self.specifications if (val := child.dump_dsl())
            ]
            if child_strings:
                return f"({op.join(child_strings)})"
        elif isinstance(self, EmptySpecification):
            return "#"
        raise ValueError(f"Unsupported specification type: {type(self)}")

    @staticmethod
    @lru_cache
    def load_dsl(dsl_string) -> Specification:
        from lark import Lark

        from fractal_specifications.generic.dsl_parser import DSLTransformer, grammar

        dsl_parser = Lark(grammar, start="start", parser="lalr")
        transformer = DSLTransformer()
        tree = dsl_parser.parse(dsl_string)
        return transformer.transform(tree)


class EmptySpecification(Specification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return True

    def to_collection(self) -> Collection:
        return []

    def __str__(self):
        return self.__class__.__name__

    def __eq__(self, other):
        return isinstance(other, EmptySpecification)

    def __hash__(self):
        return 0



from typing import Any, Collection

import pytest

from fractal_specifications.contrib.sqlalchemy.specifications import (
    SpecificationNotMappedToSqlAlchemyOrm,
    SqlAlchemyOrmSpecificationBuilder,
)


def test_build_none():
    assert SqlAlchemyOrmSpecificationBuilder.build(None) == None


def test_build_equals_specification(equals_specification):
    assert SqlAlchemyOrmSpecificationBuilder.build(equals_specification) == {"id": 1}


def test_build_or_specification(or_specification):
    assert SqlAlchemyOrmSpecificationBuilder.build(or_specification) == [
        {"id": 1},
        {"name": "test"},
    ]


def test_build_and_specification(and_specification):
    assert SqlAlchemyOrmSpecificationBuilder.build(and_specification) == {
        "id": 1,
        "name": "test",
    }


def test_build_contains_specification(contains_specification):
    with pytest.raises(SpecificationNotMappedToSqlAlchemyOrm):
        SqlAlchemyOrmSpecificationBuilder.build(contains_specification)


def test_build_in_specification(in_specification):
    with pytest.raises(SpecificationNotMappedToSqlAlchemyOrm):
        SqlAlchemyOrmSpecificationBuilder.build(in_specification)


def test_build_in_empty_specification(in_empty_specification):
    with pytest.raises(SpecificationNotMappedToSqlAlchemyOrm):
        SqlAlchemyOrmSpecificationBuilder.build(in_empty_specification)


def test_build_less_than_specification(less_than_specification):
    with pytest.raises(SpecificationNotMappedToSqlAlchemyOrm):
        SqlAlchemyOrmSpecificationBuilder.build(less_than_specification)


def test_build_less_than_equal_specification(less_than_equal_specification):
    with pytest.raises(SpecificationNotMappedToSqlAlchemyOrm):
        SqlAlchemyOrmSpecificationBuilder.build(less_than_equal_specification)


def test_build_greater_than_specification(greater_than_specification):
    with pytest.raises(SpecificationNotMappedToSqlAlchemyOrm):
        SqlAlchemyOrmSpecificationBuilder.build(greater_than_specification)


def test_build_greater_than_equal_specification(greater_than_equal_specification):
    with pytest.raises(SpecificationNotMappedToSqlAlchemyOrm):
        SqlAlchemyOrmSpecificationBuilder.build(greater_than_equal_specification)


def test_build_regex_string_match_specification(regex_string_match_specification):
    with pytest.raises(SpecificationNotMappedToSqlAlchemyOrm):
        SqlAlchemyOrmSpecificationBuilder.build(regex_string_match_specification)


def test_build_is_none_specification(is_none_specification):
    with pytest.raises(SpecificationNotMappedToSqlAlchemyOrm):
        SqlAlchemyOrmSpecificationBuilder.build(is_none_specification)


def test_build_dict_specification(dict_specification):
    assert SqlAlchemyOrmSpecificationBuilder.build(dict_specification) == {
        "id": 1,
        "test": 2,
    }


def test_build_empty_specification(empty_specification):
    assert SqlAlchemyOrmSpecificationBuilder.build(empty_specification) == None


def test_specification_not_mapped():
    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(SpecificationNotMappedToSqlAlchemyOrm):
        SqlAlchemyOrmSpecificationBuilder.build(ErrorSpecification())



import re
from functools import reduce
from typing import Callable, Dict, Iterator, Optional, Type

from django.db.models import Q  # type: ignore

from fractal_specifications.generic.specification import (
    EmptySpecification,
    Specification,
)


class SpecificationNotMappedToDjangoOrm(Exception):
    pass


class DjangoOrmSpecificationBuilder:
    @classmethod
    def build(
        cls,
        specification: Optional[Specification] = None,
    ) -> Optional[Q]:
        if specification is None:
            return None
        elif isinstance(specification, EmptySpecification):
            return None
        if builder := cls._spec_builders().get(type(specification)):
            return cls._create_q(builder(specification))
        elif isinstance(specification.to_collection(), dict):
            return cls._create_q(specification.to_collection())
        raise SpecificationNotMappedToDjangoOrm(
            f"Specification '{specification}' not mapped to Django Orm query."
        )

    @classmethod
    def _spec_builders(cls) -> Dict[Type[Specification], Callable]:
        from fractal_specifications.generic import collections, operators

        return {
            collections.AndSpecification: lambda s: reduce(
                lambda x, y: x & y, cls._build_collection(s)
            ),
            collections.OrSpecification: lambda s: reduce(
                lambda x, y: x | y, cls._build_collection(s)
            ),
            operators.EqualsSpecification: lambda s: {s.field: s.value},
            operators.InSpecification: lambda s: {f"{s.field}__in": s.value},
            operators.LessThanSpecification: lambda s: {f"{s.field}__lt": s.value},
            operators.LessThanEqualSpecification: lambda s: {
                f"{s.field}__lte": s.value
            },
            operators.GreaterThanSpecification: lambda s: {f"{s.field}__gt": s.value},
            operators.GreaterThanEqualSpecification: lambda s: {
                f"{s.field}__gte": s.value
            },
            operators.RegexStringMatchSpecification: lambda s: {
                f"{s.field}__regex": rf".*{re.escape(s.value)}.*"
            },
            operators.IsNoneSpecification: lambda s: {f"{s.field}__isnull": True},
        }

    @classmethod
    def _build_collection(cls, specification) -> Iterator[Q]:
        for spec in specification.to_collection():
            if s := cls._create_q(cls.build(spec)):
                yield s

    @staticmethod
    def _create_q(filters) -> Q:
        if type(filters) is dict:
            return Q(**filters)
        elif type(filters) in {list, set, tuple}:
            return Q(*filters)
        return filters
